""":mod:`autotweet.answer` --- Answering module
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module can auto answering to mentions to you or auto mentioning to others
tweet.

"""
import logging
import re
import time
import tweepy

from .database import NoAnswerError, get_best_answer, get_session
from .twitter import CONSUMER_KEY, CONSUMER_SECRET, OAuthToken, strip_tweet

MENTION_PATTERN = re.compile(r'(?<=\B@)\w+')
DEFAULT_THRESHOLD = 0.3

logger = logging.getLogger('answer')


def get_friends(api):
    followings = set(user.screen_name for user in api.friends())
    followers = set(user.screen_name for user in api.followers())
    friends = followings | followers
    return friends


def get_mentions(status, friends):
    user_name = status.user.screen_name
    mentions = set(MENTION_PATTERN.findall(status.text))
    mentions.discard(user_name)
    mentions = mentions & friends
    mentions = [user_name] + list(mentions)
    mentions = map(lambda x: '@' + x, mentions)
    return mentions


class MentionListener(tweepy.streaming.StreamListener):
    threshold = DEFAULT_THRESHOLD
    friends_timeout = 60
    friends_updated = None

    def __init__(self, api, db_url, threshold=None):
        super(MentionListener, self).__init__()
        self.api = api
        self.db_session = get_session(db_url)
        self.me = api.me()

        if threshold:
            self.threshold = threshold

    def get_friends(self):
        if self.friends_updated is None or \
           time.time() - self.friends_updated > self.friends_timeout:
            self.friends = get_friends(self.api)
            self.friends_updated = time.time()
            return self.friends

    def on_status(self, status):
        if hasattr(status, 'retweeted_status') or status.user.id == self.me.id:
            return True

        mentions = get_mentions(status, self.get_friends())

        question = strip_tweet(status.text)
        status_id = status.id

        try:
            answer, ratio = get_best_answer(self.db_session, question)
        except NoAnswerError:
            return True

        if status.in_reply_to_user_id == self.me.id or ratio >= self.threshold:
            logger.info(u'@{0.user.screen_name}: {0.text} -> {1}'.format(
                status, answer
            ))
            try:
                self.api.update_status(
                    u'{0} {1}'.format(' '.join(mentions), answer),
                    status_id)
            except tweepy.error.TweepError as e:
                logger.error(u'Failed to update status: {0}'.format(
                    e.message
                ))


def polling_timeline(api, db_url, threshold=None):
    db_session = get_session(db_url)
    me = api.me()
    threshold = threshold or DEFAULT_THRESHOLD

    last_id = api.home_timeline(count=1)[0].id
    logger.debug('tracking from status id: {0}'.format(last_id))

    while 1:
        time.sleep(60)
        friends = get_friends(api)

        logger.debug('polling from status id: {0}'.format(last_id))
        if not threshold:
            statuses = api.mentions_timeline(since_id=last_id)
        else:
            home_timeline = api.home_timeline(since_id=last_id)
            mentions_timeline = api.mentions_timeline(since_id=last_id)
            home_ids = [status.id for status in home_timeline]

            statuses = home_timeline + [status for status in mentions_timeline
                                        if status.id not in home_ids]

        statuses = filter(lambda x: not hasattr(x, 'retweeted_status') and
                          x.user.id != me.id,
                          statuses)

        if statuses:
            statuses.reverse()
            last_id = statuses[-1].id
        else:
            continue

        for status in statuses:
            question = strip_tweet(status.text)
            mentions = get_mentions(status, friends)

            try:
                (answer, ratio) = get_best_answer(db_session, question)
            except NoAnswerError:
                pass

            if (status.in_reply_to_user_id == me.id) or\
               (status.user.id != me.id and ratio >= threshold):
                logger.info(u'@{0.user.screen_name}: {0.text} -> {1}'.format(
                    status, answer
                ))
                try:
                    api.update_status(
                        u'{0} {1}'.format(' '.join(mentions), answer),
                        status.id)
                except tweepy.error.TweepError as e:
                    logger.error(u'Failed to update status: {0}'.format(
                        e.message
                    ))


def answer_daemon(token, db_url, streaming=False, threshold=None):
    if not isinstance(token, OAuthToken):
        token = OAuthToken.from_string(token)

    auth = tweepy.OAuthHandler(CONSUMER_KEY, CONSUMER_SECRET)
    auth.set_access_token(token.key, token.secret)
    api = tweepy.API(auth)

    if streaming:
        listener = MentionListener(api, db_url, threshold=threshold)

        stream = tweepy.Stream(auth, listener)
        stream.userstream()
    else:
        polling_timeline(api, db_url, threshold)



""":mod:`autotweet.twitter` --- Twitter utilities
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains Twitter API key and some useful methods.

"""
import cgi
import re
import tweepy
import webbrowser
try:
    from HTMLParser import HTMLParser
except ImportError:
    from html.parser import HTMLParser
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode


__all__ = ('CONSUMER_KEY', 'CONSUMER_SECRET', 'authorize', 'strip_tweet')


#: Consumer key for autoweet.
CONSUMER_KEY = '62yWrV2RhpGgWOKlqvJPNQ'
#: Consumer secret key for autotweet.
CONSUMER_SECRET = 'Je6NLI7AN3c1BJP9kHaq1p8GBkMyKs5GhX954dWJ6I'

url_pattern = re.compile(r'https?://[^\s]+')
mention_pattern = re.compile(r'@\w+')
html_parser = HTMLParser()


class OAuthToken(object):
    key = None
    secret = None

    def __init__(self, key, secret):
        self.key = key
        self.secret = secret

    def to_string(self):
        return urlencode({
            'oauth_token': self.key,
            'oauth_token_secret': self.secret,
        })

    @staticmethod
    def from_string(string):
        params = cgi.parse_qs(string, keep_blank_values=False)
        key = params['oauth_token'][0]
        secret = params['oauth_token_secret'][0]
        return OAuthToken(key, secret)


def authorize():
    """Authorize to twitter.

    Use PIN authentification.

    :returns: Token for authentificate with Twitter.
    :rtype: :class:`autotweet.twitter.OAuthToken`

    """
    auth = tweepy.OAuthHandler(CONSUMER_KEY, CONSUMER_SECRET)
    url = auth.get_authorization_url()
    print('Open this url on your webbrowser: {0}'.format(url))
    webbrowser.open(url)
    pin = raw_input('Input verification number here: ').strip()

    token_key, token_secret = auth.get_access_token(verifier=pin)

    return OAuthToken(token_key, token_secret)


def strip_tweet(text, remove_url=True):
    """Strip tweet message.

    This method removes mentions strings and urls(optional).

    :param text: tweet message
    :type text: :class:`str`

    :param remove_url: Remove urls. default :const:`True`.
    :type remove_url: :class:`boolean`

    :returns: Striped tweet message
    :rtype: :class:`str`

    """
    if remove_url:
        text = url_pattern.sub('', text)
    text = mention_pattern.sub('', text)
    text = html_parser.unescape(text)
    text = text.strip()
    return text



""":mod:`autotweet.database` --- Database structure
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides methods to get session, get answer, etc.

"""
import logging
import math
import random
from sqlalchemy import (Column, Float,  ForeignKey, Integer, String, Table,
                        create_engine)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, scoped_session, sessionmaker


__all__ = ('Document', 'Gram', 'NoAnswerError', 'get_count', 'get_session',
           'add_document', 'get_best_answer', 'recreate_grams', 'recalc_idfs')


Base = declarative_base()
GRAM_LENGTH = 2

logger = logging.getLogger('database')


association_table = Table(
    'association', Base.metadata,
    Column('gram_id', Integer, ForeignKey('grams.id')),
    Column('document_id', Integer, ForeignKey('documents.id'))
    )


class Document(Base):
    __tablename__ = 'documents'
    id = Column(Integer, primary_key=True)
    text = Column(String(140), nullable=False)
    answer = Column(String(140), nullable=False)
    grams = relationship(
        'Gram', secondary=association_table, backref='documents')

    def __init__(self, text, answer):
        self.text = text
        self.answer = answer


class Gram(Base):
    __tablename__ = 'grams'
    id = Column(Integer, primary_key=True)
    gram = Column(String(GRAM_LENGTH), unique=True, nullable=False)
    idf = Column(Float)

    def __init__(self, gram):
        self.gram = gram


def get_session(url):
    """Get db session.

    :param url: URL for connect with DB
    :type url: :class:`str`

    :returns: A sqlalchemy db session
    :rtype: :class:`sqlalchemy.orm.Session`

    """
    engine = create_engine(url)
    db_session = scoped_session(
        sessionmaker(engine, autoflush=True, autocommit=True))
    Base.metadata.create_all(engine)
    Base.query = db_session.query_property()
    return db_session


def add_document(session, question, answer):
    """Add question answer set to DB.

    :param session: DB session
    :type session: :class:`sqlalchemt.orm.Session`

    :param question: A question to an answer
    :type question: :class:`str`

    :param answer: An answer to a question
    :type answer: :class:`str`

    """
    question = question.strip()
    answer = answer.strip()

    if session.query(Document)\
            .filter_by(text=question, answer=answer).count():
        logger.info(u'Already here: {0} -> {1}'.format(question, answer))
        return
    logger.info(u'add document: {0} -> {1}'.format(question, answer))

    session.begin()

    grams = _get_grams(session, question, make=True)

    doc = Document(question, answer)
    doc.grams = list(grams)
    recalc_idfs(session, grams)

    session.add(doc)
    session.commit()


def get_best_answer(session, query):
    """Get best answer to a question.

    :param session: DB session
    :type session: :class:`sqlalchemt.orm.Session`

    :param query: A question to get an answer
    :type query: :class:`str`

    :returns: An answer to a question
    :rtype: :class:`str`

    :raises: :class:`NoAnswerError` when can not found answer to a question

    """
    if not isinstance(query, unicode):
        query = query.decode('utf-8')

    session.begin()

    grams = _get_grams(session, query)
    if not grams:
        raise NoAnswerError('Can not found answer')

    documents = set([doc for gram in grams for doc in gram.documents])

    recalc_idfs(session, grams)

    idfs = dict((gram.gram, gram.idf) for gram in grams)

    docs = dict(
        (doc.answer, _cosine_measure(idfs, _get_tf_idfs(session, doc)))
        for doc in documents)
    docs = dict((key, val) for (key, val) in docs.items() if val)

    session.commit()

    try:
        max_ratio = max(docs.values())
        answers = [answer for answer in docs.keys()
                   if docs.get(answer) == max_ratio]

        answer = random.choice(answers)
        logger.debug(u'{0} -> {1} ({2})'.format(query, answer, max_ratio))
        return (answer, max_ratio)
    except ValueError:
        raise NoAnswerError('Can not found answer')


def recreate_grams(session):
    """Re-create grams for database.

    In normal situations, you never need to call this method.
    But after migrate DB, this method is useful.

    :param session: DB session
    :type session: :class:`sqlalchemt.orm.Session`

    """
    session.begin()

    for document in session.query(Document).all():
        grams = _get_grams(session, document.text, make=True)
        document.grams = list(grams)

    broken_links = session.query(Gram)\
        .filter(~Gram.documents.any()).all()
    for gram in broken_links:
        session.delete(gram)

    session.commit()


def recalc_idfs(session, grams=None):
    """Re-calculate idfs for database.

    calculating idfs for gram is taking long time.
    So I made it calculates idfs for some grams.
    If you want make accuracy higher, use this with grams=None.

    :param session: DB session
    :type session: :class:`sqlalchemt.orm.Session`

    :param grams: grams that you want to re-calculating idfs
    :type grams: A set of :class:`Gram`

    """
    session.begin(subtransactions=True)

    if not grams:
        grams = session.query(Gram).all()
    for gram in grams:
        gram.idf = _get_idf(session, gram)

    session.commit()


def get_count(session):
    """Get count of :class:`Document`.

    :param session: DB session
    :type session: :class:`sqlalchemt.orm.Session`

    """
    return session.query(Document).count()


def _get_grams(session, text, make=False):
    grams = set()
    session.begin(subtransactions=True)

    if len(text) < GRAM_LENGTH:
        gram_obj = session.query(Gram).filter_by(gram=text).first()
        if gram_obj:
            grams.add(gram_obj)
        elif make:
            gram_obj = Gram(text)
            session.add(gram_obj)
            grams.add(gram_obj)
    else:
        for i in range(len(text) - GRAM_LENGTH + 1):
            gram = text[i:i+GRAM_LENGTH]
            gram_obj = session.query(Gram).filter_by(gram=gram).first()
            if gram_obj:
                grams.add(gram_obj)
            elif make:
                gram_obj = Gram(gram)
                session.add(gram_obj)
                grams.add(gram_obj)

    session.commit()
    return grams


def _get_tf(gram, document):
    if isinstance(gram, Gram):
        gram = gram.gram

    if not isinstance(gram, unicode):
        gram = gram.decode('utf-8')

    return document.text.count(gram) + document.answer.count(gram)


def _get_idf(session, gram):
    all_count = session.query(Document).count()
    d_count = len(gram.documents)
    return math.log((all_count / (1.0 + d_count)) + 1)


def _get_tf_idfs(self, document):
    tf_idfs = dict(
        (gram.gram, _get_tf(gram, document) * gram.idf)
        for gram in document.grams
        if gram.idf is not None)
    return tf_idfs


def _cosine_measure(v1, v2):
    intersection = set(v1.keys()) & set(v2.keys())
    numerator = sum([v1[x] * v2[x] for x in intersection])

    sum1 = sum([v1[x]**2 for x in v1.keys() if v1[x] is not None])
    sum2 = sum([v2[y]**2 for y in v2.keys() if v2[y] is not None])

    denominator = math.sqrt(sum1) * math.sqrt(sum2)

    try:
        return numerator / denominator
    except ZeroDivisionError:
        return 0


class NoAnswerError(Exception):
    """Raises when autotweet can not found best answer to a question.

    :param msg: A message for the exception
    :type msg: :class:`str`

    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)



""":mod:`autotweet.learn` --- Learning your tweets
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module learns your tweets and store it to database.

"""
import logging
import sqlalchemy
import time
import tweepy
from .database import add_document, get_session
from .twitter import CONSUMER_KEY, CONSUMER_SECRET, OAuthToken, strip_tweet

MY_CLIENT_NAME = 'learn your tweet'
logger = logging.getLogger('collector')


class MyMentionListener(tweepy.streaming.StreamListener):

    def __init__(self, api, db_url):
        super(MyMentionListener, self).__init__()
        self.api = api
        self.db_url = db_url
        self.db_session = get_session(db_url)
        self.me = api.me()

    def on_status(self, status):
        if hasattr(status, 'retweeted_status'):
            return True

        if status.user.id == self.me.id and status.in_reply_to_status_id:
            if status.source == MY_CLIENT_NAME:
                return True
            original_status = self.api.get_status(status.in_reply_to_status_id)

            question = strip_tweet(original_status.text)
            answer = strip_tweet(status.text, remove_url=False)

            if question and answer:
                try:
                    add_document(self.db_session, question, answer)
                except sqlalchemy.exc.OperationalError:
                    self.db_session = get_session(self.db_url)
                    add_document(self.db_session, question, answer)

        return True


def polling_timeline(api, db_url):
    db_session = get_session(db_url)
    me = api.me()
    last_id = me.status.id

    logger.debug('tracking from status id: {0}'.format(last_id))
    while 1:
        time.sleep(60)
        logger.debug('polling from status id: {0}'.format(last_id))
        statuses = me.timeline(since_id=last_id)
        if statuses:
            statuses.reverse()
            last_id = statuses[-1].id
        else:
            continue

        for status in statuses:
            if status.source == MY_CLIENT_NAME:
                continue
            if hasattr(status, 'retweeted_status'):
                continue
            if not status.in_reply_to_status_id:
                continue

            original_status = api.get_status(status.in_reply_to_status_id)

            question = strip_tweet(original_status.text)
            answer = strip_tweet(status.text, remove_url=False)

            if question and answer:
                try:
                    add_document(db_session, question, answer)
                except sqlalchemy.exc.OperationalError:
                    db_session = get_session(db_url)
                    add_document(db_session, question, answer)


def learning_daemon(token, db_url, streaming=False):
    if not isinstance(token, OAuthToken):
        token = OAuthToken.from_string(token)

    auth = tweepy.OAuthHandler(CONSUMER_KEY, CONSUMER_SECRET)
    auth.set_access_token(token.key, token.secret)
    api = tweepy.API(auth)

    if streaming:
        listener = MyMentionListener(api, db_url)

        stream = tweepy.Stream(auth, listener)
        stream.userstream()
    else:
        polling_timeline(api, db_url)
